use anyhow::Error;
use anyhow::{anyhow, Result};

use crate::graph_types::*;
use crate::id_types::*;

impl Graph {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn add_node(&mut self, d: NodeDescriptor) -> NodeId {
        let node_id = self.nodes.insert_with_key(|node_id| {
            Node {
                id: node_id,
                op_name: d.op_name,
                // These will get filled in later
                inputs: Vec::default(),
                outputs: Vec::default(),
            }
        });

        let inputs: Vec<(String, ID)> = d
            .inputs
            .into_iter()
            .map(|(input_name, input)| {
                let input_id = self.inputs.insert_with_key(|id| match input {
                    Descriptor::Button => Param {
                        id,
                        kind: input,
                        node: node_id,
                        draw_connector: true,
                    },
                    Descriptor::Slider { .. } => Param {
                        id,
                        kind: input,
                        node: node_id,
                        draw_connector: true,
                    },
                    Descriptor::Info => Param {
                        id,
                        kind: input,
                        node: node_id,
                        draw_connector: false,
                    },
                });

                (input_name, input_id)
            })
            .collect();

        let outputs: Vec<(String, ID)> = d
            .outputs
            .into_iter()
            .map(|(output_name, output)| {
                let output_id = self.outputs.insert_with_key(|id| match output {
                    Descriptor::Button => Param {
                        id,
                        kind: output,
                        node: node_id,
                        draw_connector: true,
                    },
                    Descriptor::Slider { .. } => Param {
                        id,
                        kind: output,
                        node: node_id,
                        draw_connector: true,
                    },
                    Descriptor::Info => Param {
                        id,
                        kind: output,
                        node: node_id,
                        draw_connector: false,
                    },
                });
                (output_name, output_id)
            })
            .collect();

        self[node_id].inputs = inputs;
        self[node_id].outputs = outputs;
        node_id
    }

    pub fn remove_node(&mut self, node_id: NodeId) {
        self.connections
            .retain(|i, o| !(self.outputs[*o].node == node_id || self.inputs[*i].node == node_id));
        let inputs: Vec<_> = self[node_id].input_ids().collect();
        for input in inputs {
            self.inputs.remove(input);
        }
        let outputs: Vec<_> = self[node_id].output_ids().collect();
        for output in outputs {
            self.outputs.remove(output);
        }
        self.nodes.remove(node_id);
    }

    pub fn remove_connection(&mut self, input_id: ID) -> Option<ID> {
        self.connections.remove(&input_id)
    }

    pub fn iter_nodes(&self) -> impl Iterator<Item = NodeId> + '_ {
        self.nodes.iter().map(|(id, _)| id)
    }

    pub fn add_connection(&mut self, output: ID, input: ID) {
        self.connections.insert(input, output);
    }

    pub fn iter_connections(&self) -> impl Iterator<Item = (ID, ID)> + '_ {
        self.connections.iter().map(|(o, i)| (*o, *i))
    }

    pub fn connection(&self, input: ID) -> Option<ID> {
        self.connections.get(&input).copied()
    }

    pub fn any_type(&self, id: ID) -> Result<&Param> {
        match self.inputs.get(id) {
            Some(param) => Ok(param),
            None => match self.outputs.get(id) {
                Some(param) => Ok(param),
                None => Err(Error::msg(format!("Invalid parameter id: {:?}", id))),
            },
        }
    }
}

impl Node {
    pub fn inputs<'a>(&'a self, graph: &'a Graph) -> impl Iterator<Item = &Param> + 'a {
        self.input_ids().map(|id| graph.input(id))
    }

    pub fn outputs<'a>(&'a self, graph: &'a Graph) -> impl Iterator<Item = &Param> + 'a {
        self.output_ids().map(|id| graph.output(id))
    }

    pub fn input_ids(&self) -> impl Iterator<Item = ID> + '_ {
        self.inputs.iter().map(|(_name, id)| *id)
    }

    pub fn output_ids(&self) -> impl Iterator<Item = ID> + '_ {
        self.outputs.iter().map(|(_name, id)| *id)
    }

    pub fn get_input(&self, name: &str) -> Result<ID> {
        self.inputs
            .iter()
            .find(|(param_name, _id)| param_name == name)
            .map(|x| x.1)
            .ok_or_else(|| anyhow!("Node {:?} has no parameter named {}", self.id, name))
    }

    pub fn get_output(&self, name: &str) -> Result<ID> {
        self.outputs
            .iter()
            .find(|(param_name, _id)| param_name == name)
            .map(|x| x.1)
            .ok_or_else(|| anyhow!("Node {:?} has no parameter named {}", self.id, name))
    }

    // Can this node be enabled on the UI? I.e. does it output a mesh?
    // pub fn can_be_enabled(&self, graph: &Graph) -> bool {
    //     self.outputs(graph)
    //         .any(|output| output.typ == DataType::Mesh)
    // }
}
